"""
Benchmark functions for imports
"""


def timeraw_import_locan():
    """Benchmark import time for locan"""
    return """
    import locan
    """


def timeraw_import_locan_analysis():
    """Benchmark import time for locan"""
    return """
    import locan.analysis
    """


def timeraw_import_locan_configuration():
    """Benchmark import time for locan"""
    return """
    import locan.configuration
    """


def timeraw_import_locan_constants():
    """Benchmark import time for locan"""
    return """
    import locan.constants
    """


def timeraw_import_locan_data():
    """Benchmark import time for locan"""
    return """
    import locan.data
    """


def timeraw_import_locan_datasets():
    """Benchmark import time for locan"""
    return """
    import locan.datasets
    """


def timeraw_import_locan_dependencies():
    """Benchmark import time for locan"""
    return """
    import locan.dependencies
    """


def timeraw_import_locan_gui():
    """Benchmark import time for locan"""
    return """
    import locan.gui
    """


def timeraw_import_locan_locan_io():
    """Benchmark import time for locan"""
    return """
    import locan.locan_io
    """


def timeraw_import_locan_rois():
    """Benchmark import time for locan"""
    return """
    import locan.rois
    """


def timeraw_import_locan_scripts():
    """Benchmark import time for locan"""
    return """
    import locan.scripts
    """


def timeraw_import_locan_simulation():
    """Benchmark import time for locan"""
    return """
    import locan.simulation
    """


def timeraw_import_locan_tests():
    """Benchmark import time for locan"""
    return """
    import locan.tests
    """


def timeraw_import_locan_utils():
    """Benchmark import time for locan"""
    return """
    import locan.utils
    """


def timeraw_import_locan_visualize():
    """Benchmark import time for locan"""
    return """
    import locan.visualize
    """


def timeraw_import_asdf():
    """Benchmark import time"""
    return """
    import asdf
    """


def timeraw_import_boost_histogram():
    """Benchmark import time"""
    return """
    import boost_histogram
    """


def timeraw_import_fast_histogram():
    """Benchmark import time"""
    return """
    import fast_histogram
    """


def timeraw_import_lmfit():
    """Benchmark import time"""
    return """
    import lmfit
    """


def timeraw_import_matplotlib_pyplot():
    """Benchmark import time"""
    return """
    import matplotlib.pyplot
    """


def timeraw_import_networkx():
    """Benchmark import time"""
    return """
    import networkx
    """


def timeraw_import_numpy():
    """Benchmark import time for numpy"""
    return """
    import numpy
    """


def timeraw_import_pandas():
    """Benchmark import time for pandas"""
    return """
    import pandas
    """


def timeraw_import_protobuf():
    """Benchmark import time"""
    return """
    import google.protobuf
    """


def timeraw_import_qtpy():
    """Benchmark import time"""
    return """
    import qtpy
    """


def timeraw_import_ruamel_yaml():
    """Benchmark import time"""
    return """
    import ruamel.yaml
    """


def timeraw_import_skimage():
    """Benchmark import time"""
    return """
    import skimage
    """


def timeraw_import_sklearn():
    """Benchmark import time"""
    return """
    import sklearn
    """


def timeraw_import_scipy():
    """Benchmark import time"""
    return """
    import scipy
    """


def timeraw_import_shapely():
    """Benchmark import time"""
    return """
    import shapely
    """


def timeraw_import_tifffile():
    """Benchmark import time"""
    return """
    import tifffile
    """


def timeraw_import_tqdm():
    """Benchmark import time"""
    return """
    import tqdm
    """
