# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: locan/locan_io/locdata/manifest.proto
# Protobuf Python Version: 5.28.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    2,
    "",
    "locan/locan_io/locdata/manifest.proto",
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from locan.data import metadata_pb2 as locan_dot_data_dot_metadata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%locan/locan_io/locdata/manifest.proto\x12\x16locan.locan_io.locdata\x1a\x19locan/data/metadata.proto"\xf1\x01\n\x06\x46ormat\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Type\x12*\n\x04mode\x18\x03 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Mode\x12\x0f\n\x07\x63olumns\x18\x04 \x01(\x03\x12\x0f\n\x07headers\x18\x05 \x03(\t\x12,\n\x05\x64type\x18\x06 \x03(\x0e\x32\x1d.locan.locan_io.locdata.Dtype\x12\r\n\x05shape\x18\x07 \x03(\x03\x12\r\n\x05units\x18\x08 \x03(\t\x12\x13\n\x0b\x64\x65scription\x18\t \x01(\t"\xba\x03\n\x08\x46ileInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Type\x12\x0e\n\x06\x66ormat\x18\x03 \x01(\t\x12\x0f\n\x07\x63hannel\x18\x04 \x01(\t\x12\x0c\n\x04rows\x18\x05 \x01(\x03\x12<\n\x06offset\x18\x06 \x03(\x0b\x32,.locan.locan_io.locdata.FileInfo.OffsetEntry\x12\x36\n\x03min\x18\x07 \x03(\x0b\x32).locan.locan_io.locdata.FileInfo.MinEntry\x12\x36\n\x03max\x18\x08 \x03(\x0b\x32).locan.locan_io.locdata.FileInfo.MaxEntry\x12\x10\n\x08\x65xposure\x18\t \x01(\x02\x1a-\n\x0bOffsetEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a*\n\x08MinEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a*\n\x08MaxEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01"\x98\x04\n\x08Manifest\x12\x16\n\x0e\x66ormat_version\x18\x01 \x01(\t\x12>\n\x07\x66ormats\x18\x02 \x03(\x0b\x32-.locan.locan_io.locdata.Manifest.FormatsEntry\x12/\n\x05\x66iles\x18\x03 \x03(\x0b\x32 .locan.locan_io.locdata.FileInfo\x12\x0c\n\x04name\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x11\n\tthumbnail\x18\x07 \x01(\t\x12\x0e\n\x06sample\x18\x08 \x01(\t\x12\x10\n\x08labeling\x18\t \x01(\t\x12\x0c\n\x04\x64\x61te\x18\n \x01(\t\x12\x0e\n\x06\x61uthor\x18\x0b \x01(\t\x12\x10\n\x08\x63itation\x18\x0c \x01(\t\x12\r\n\x05\x65mail\x18\r \x01(\t\x12\x36\n\x03map\x18\x0e \x03(\x0b\x32).locan.locan_io.locdata.Manifest.MapEntry\x12*\n\x0clocdata_meta\x18\x0f \x01(\x0b\x32\x14.locan.data.Metadata\x1aN\n\x0c\x46ormatsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.locan.locan_io.locdata.Format:\x02\x38\x01\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*\x1c\n\x04Type\x12\t\n\x05TABLE\x10\x00\x12\t\n\x05IMAGE\x10\x01*\x1c\n\x04Mode\x12\n\n\x06\x42INARY\x10\x00\x12\x08\n\x04TEXT\x10\x01*{\n\x05\x44type\x12\x08\n\x04INT8\x10\x01\x12\t\n\x05UINT8\x10\x02\x12\t\n\x05INT16\x10\x03\x12\n\n\x06UINT16\x10\x04\x12\t\n\x05INT32\x10\x05\x12\n\n\x06UINT32\x10\x06\x12\t\n\x05INT64\x10\x07\x12\n\n\x06UINT64\x10\x08\x12\x0b\n\x07\x46LOAT32\x10\t\x12\x0b\n\x07\x46LOAT64\x10\n'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "locan.locan_io.locdata.manifest_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_FILEINFO_OFFSETENTRY"]._loaded_options = None
    _globals["_FILEINFO_OFFSETENTRY"]._serialized_options = b"8\001"
    _globals["_FILEINFO_MINENTRY"]._loaded_options = None
    _globals["_FILEINFO_MINENTRY"]._serialized_options = b"8\001"
    _globals["_FILEINFO_MAXENTRY"]._loaded_options = None
    _globals["_FILEINFO_MAXENTRY"]._serialized_options = b"8\001"
    _globals["_MANIFEST_FORMATSENTRY"]._loaded_options = None
    _globals["_MANIFEST_FORMATSENTRY"]._serialized_options = b"8\001"
    _globals["_MANIFEST_MAPENTRY"]._loaded_options = None
    _globals["_MANIFEST_MAPENTRY"]._serialized_options = b"8\001"
    _globals["_TYPE"]._serialized_start = 1320
    _globals["_TYPE"]._serialized_end = 1348
    _globals["_MODE"]._serialized_start = 1350
    _globals["_MODE"]._serialized_end = 1378
    _globals["_DTYPE"]._serialized_start = 1380
    _globals["_DTYPE"]._serialized_end = 1503
    _globals["_FORMAT"]._serialized_start = 93
    _globals["_FORMAT"]._serialized_end = 334
    _globals["_FILEINFO"]._serialized_start = 337
    _globals["_FILEINFO"]._serialized_end = 779
    _globals["_FILEINFO_OFFSETENTRY"]._serialized_start = 646
    _globals["_FILEINFO_OFFSETENTRY"]._serialized_end = 691
    _globals["_FILEINFO_MINENTRY"]._serialized_start = 693
    _globals["_FILEINFO_MINENTRY"]._serialized_end = 735
    _globals["_FILEINFO_MAXENTRY"]._serialized_start = 737
    _globals["_FILEINFO_MAXENTRY"]._serialized_end = 779
    _globals["_MANIFEST"]._serialized_start = 782
    _globals["_MANIFEST"]._serialized_end = 1318
    _globals["_MANIFEST_FORMATSENTRY"]._serialized_start = 1196
    _globals["_MANIFEST_FORMATSENTRY"]._serialized_end = 1274
    _globals["_MANIFEST_MAPENTRY"]._serialized_start = 1276
    _globals["_MANIFEST_MAPENTRY"]._serialized_end = 1318
# @@protoc_insertion_point(module_scope)
